/************************************************************
*************************************************************
*                                                           *
* All relevant (system) settings are done in configurations.h *
*                                                           *
*************************************************************
************************************************************/


/****************************************************************************
 Include Section
*****************************************************************************/

#include "configurations.h"
#if defined DATA_CACHE_EN

#include <cplb.h>
#include <def_LPBlackfin.h>
#include "macros.h"


/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __dcplb_init;
.GLOBAL __dcplb_handler;


/*****************************************************************************
 Global Variables
******************************************************************************/

.GLOBAL _b4_dcplb_data_pointer;
.GLOBAL _b4_dcplb_address_pointer;


/*****************************************************************************
 Variables
******************************************************************************/

.SECTION cplb_data;

.ALIGN 4;
.BYTE4 _b4_dcplb_data_pointer = 0;
.BYTE4 _b4_dcplb_address_pointer = 0;


/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

/* L1 Data Cache Bank Select */
/* Standard definition does not work well */
#define         CPLB_DCBS               0x10


/* number of entries in _b4_dcplbs_table */
#if SDRAM_SIZE == 32
    #if defined __ADSPBF548__
    #define     dcplbs_table_length     11
    #else
    #define     dcplbs_table_length     10
    #endif
#endif

#if SDRAM_SIZE == 64
    #if defined __ADSPBF548__
    #define     dcplbs_table_length     19
    #else
    #define     dcplbs_table_length     18
    #endif
#endif

#if (SDRAM_SIZE == 128) || defined (LIMIT32kB)
    #undef dcplbs_table_length
    #if defined __ADSPBF548__
    #define     dcplbs_table_length     35
    #else
    #define     dcplbs_table_length     34
    #endif
#endif


/* number of entries in _b4_dcplbs_table is depending on the PAGE_SIZE */
#define         PAGE_SIZE               PAGE_SIZE_4MB

#if defined LIMIT32kB
    #undef      PAGE_SIZE
    #define     PAGE_SIZE               PAGE_SIZE_1KB
#endif


/***************************************************************************************
L1 Data Memory Control Register (DMEM_CONTROL)
RESERVED        [0]     1
ENDCPLB         [1]     Data CPLB enable
                        0 - CLPBs disabled, minimal address checking only
 CPLB_ENABLE_DCPLBS     1 - CPLBs enabled
DMC           [3:2]     Enable/Disable Data Cache functionality
        ASRAM_BSRAM     00 - Both data banks are SRAM, also invalidates all cache lines if previously configured as cache
                        01 - Reserved
       ACACHE_BSRAM     10 - Data bank A is lower 16K byte SRAM, upper 16K byte cache Data bank B is SRAM
      ACACHE_BCACHE     11 - Both data banks are lower 16K byte SRAM, upper 16K byte cache
RESERVED    [11: 5]
DCBS            [4]     L1 Data Cache Bank Select
                        Valid only when DMC[1:0] = 11. Determines whether Address bit A[14] or A[23] is used to select the L1 data cache bank
                        0 - A[14] (16kB switch)
               DCBS     1 - A[23] (8MB switch)
PORT_PREF0     [12]     DAG0 Port Preference
                        0 - DAG0 non-cacheable fetches use port A
         PORT_PREF0     1 - DAG0 non-cacheable fetches use port B
PORT_PREF1     [13]     DAG1 Port Preference
                        0 - DAG1 non-cacheable fetches use port A
         PORT_PREF1     1 - DAG1 non-cacheable fetches use port B
RESERVED    [31:14]
***************************************************************************************/

#define SET_BIT_0               0x1 /* Preserve Reset Status */

#if defined (DATA_CACHE_AB) && defined (DATA_CACHE_DCBS)
#define DMEM_CONTROL_VAL        ( ENDCPLB | ACACHE_BCACHE | CPLB_DCBS | PORT_PREF0 | SET_BIT_0 )
#endif

#if defined (DATA_CACHE_AB) && !defined (DATA_CACHE_DCBS)
#define DMEM_CONTROL_VAL        ( ENDCPLB | ACACHE_BCACHE             | PORT_PREF0 | SET_BIT_0 )
#endif

#if !defined DATA_CACHE_AB
#define DMEM_CONTROL_VAL        ( ENDCPLB | ACACHE_BSRAM              | PORT_PREF0 | SET_BIT_0 )
#endif


/***************************************************************************************
DCPLB Data Registers (DCPLB_DATAx)
CPLB_VALID              [0]
                                0 - Invalid (disabled) CPLB entry
                 CPLB_VALID     1 - Valid (enabled) CPLB entry
CPLB_LOCK               [1]     Can be used by software in CPLB replacement algorithms
                                0 - Unlocked, CPLB entry can be replaced
                  CPLB_LOCK     1 - Locked, CPLB entry should not be replaced
CPLB_USER_RD            [2]
                                0 - User mode read access generates protection violation exception
               CPLB_USER_RD     1 - User mode read access permitted
CPLB_USER_WR            [3]
                                0 - User mode write access generates protection violation exception
               CPLB_USER_WR     1 - User mode write access permitted
CPLB_SUPV_WR            [4]
                                0 - Supervisor mode write access generates protection violation exception
               CPLB_SUPV_WR     1 - Supervisor mode write access permitted
RESERVED            [ 6: 5]
CPLB_DIRTY              [7]     Valid only if write back cacheable (CPLB_VALID = 1, CPLB_WT = 0, and CPLB_L1_CHBL = 1)
                                0 - Clean
                                1 - Dirty
                                A protection violation exception is generated on store accesses to this page when this bit is 0.
                                The state of this bit is modified only by writes to this register.
                                The exception service routine must set this bit.
RESERVED            [11: 8]
CPLB_L1_CHBL           [12]     Clear this bit whenever L1 memory is configured as SRAM
                                0 - Non-cacheable in L1
                                1 - Cacheable in L1
RESERVED               [13]
CPLB_WT                [14]     Operates only in cache mode
                                0 - Write back
                    CPLB_WT     1 - Write through
CPLB_A1_AOW            [15]     Valid only if write through cacheable (CPLB_VALID = 1, CPLB_WT = 1)
                                0 - Allocate cache lines on reads only
                CPLB_L1_AOW     1 - Allocate cache lines on reads and writes
PAGE_SIZE           [17:16]
              PAGE_SIZE_1KB     00 - 1K byte page size
              PAGE_SIZE_4KB     01 - 4K byte page size
              PAGE_SIZE_1MB     10 - 1M byte page size
              PAGE_SIZE_4MB     11 - 4M byte page size
RESERVED            [31:18]
***************************************************************************************/

#define DCPLB_DATA_L1_VAL                       ( CPLB_VALID | CPLB_LOCK | CPLB_SUPV_WR                | PAGE_SIZE_4MB )

#if defined (DATA_CACHE_WT) && !defined (DATA_CACHE_AOW)
    #define DCPLB_DATA_SDRAM_CACHE_VAL          ( CPLB_VALID | CPLB_USER_RD | CPLB_USER_WR | CPLB_SUPV_WR              | CPLB_L1_CHBL | CPLB_WT | PAGE_SIZE )
#endif

#if defined (DATA_CACHE_WT) && defined (DATA_CACHE_AOW)
    #define DCPLB_DATA_SDRAM_CACHE_VAL          ( CPLB_VALID | CPLB_USER_RD | CPLB_USER_WR | CPLB_SUPV_WR              | CPLB_L1_CHBL | CPLB_WT | CPLB_L1_AOW | PAGE_SIZE )
#endif

#if !defined (DATA_CACHE_WT)
    #define DCPLB_DATA_SDRAM_CACHE_VAL          ( CPLB_VALID | CPLB_USER_RD | CPLB_USER_WR | CPLB_SUPV_WR | CPLB_DIRTY | CPLB_L1_CHBL | PAGE_SIZE )
#endif

#if defined __ADSPBF548__
    #define DCPLB_DATA_L2_SRAM_VAL              ( CPLB_VALID | CPLB_USER_RD | CPLB_USER_WR | CPLB_SUPV_WR                             | PAGE_SIZE_1MB )
#endif

#define DCPLB_DATA_ASYNC_CACHE_VAL              ( CPLB_VALID | CPLB_LOCK                | CPLB_L1_CHBL | PAGE_SIZE_4MB )
#define DCPLB_DATA_ASYNC_NOCACHE_VAL            ( CPLB_VALID | CPLB_LOCK                               | PAGE_SIZE_4MB )

#define DCPLB_DATA_SDRAM_NOCACHE_VAL            ( CPLB_VALID | CPLB_USER_RD | CPLB_USER_WR | CPLB_SUPV_WR |                           | PAGE_SIZE )


/*****************************************************************************
 DCPLB Data
******************************************************************************/

.SECTION cplb_data;
.ALIGN 4;


_b4_dcplbs_table:

.BYTE4=

//    0xFFC00000, (DCPLB_DATA_L1_VAL),            /* CORE/SYSTEM MMR */
    0xFF800000, (DCPLB_DATA_L1_VAL),            /* L1 SRAM */

#if defined __ADSPBF548__
    0xFEB00000, (DCPLB_DATA_L2_SRAM_VAL),       /* L2 SRAM */
#endif

    0x20000000, (DCPLB_DATA_ASYNC_NOCACHE_VAL), /* Async Memory Bank 0-3 */


#if ( SDRAM_SIZE == 64 || SDRAM_SIZE == 32 ) && !defined ( LIMIT32kB )

    0x00000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 16MB */

#endif


#if ( SDRAM_SIZE == 64 || SDRAM_SIZE == 32 ) && !defined ( LIMIT32kB )

    0x01000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x01400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x01800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x01C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 32MB */

#endif


#if ( SDRAM_SIZE == 64 ) && !defined ( LIMIT32kB )

    0x02000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x02400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x02800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x02C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x03000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x03400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x03800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x03C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 64MB */

#endif


#if ( SDRAM_SIZE == 128 ) && !defined ( LIMIT32kB )

    0x04000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x04400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x04800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x04C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x05000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x05400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x05800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x05C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x06000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x06400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x06800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x06C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x07000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x07400000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x07800000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x07C00000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 128MB */

#endif


#if defined LIMIT32kB

    0x00000000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00000400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00000800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00000C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00001000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00001400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00001800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00001C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 8kB */

    0x00002000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00002400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00002800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00002C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00003000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00003400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00003800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00003C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 16kB */
    
    0x00004000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00004400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00004800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00004C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00005000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00005400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00005800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00005C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 24kB */
    
    0x00006000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00006400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00006800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00006C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00007000,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00007400,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00007800,(DCPLB_DATA_SDRAM_CACHE_VAL),
    0x00007C00,(DCPLB_DATA_SDRAM_CACHE_VAL),
    /* --> 32kB */

#endif


    /* end of section - termination */
    0xFFFFFFFF;

_b4_dcplbs_table.END:


/*****************************************************************************
 DCPLB Code
******************************************************************************/

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION cplb_code;


/****************************************************************************
 Function:      __dcplb_init
 
 Description:   Initialize L1 Data Memory Registers
                according to the _b4_dcplbs_table
*****************************************************************************/


__dcplb_init:


    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:5,P5:4);
    PUSH_DAG_REGx(0);
    PUSH_DAG_REGx(1);
    PUSH_DAG_REGx(2);
    IMM32(P5,SYS_MMR_BASE);


    M0 = 0(z); M1 = 0(z); M2 = 0(z);
    L0 = 0(z); L1 = 0(z); L2 = 0(z);
    B0 = 0(z); B1 = 0(z); B2 = 0(z);

    /* RESET DMEM_CONTROL register */
    R7 = 0x1 (z);
    [P5 + lo(DMEM_CONTROL)] = R7;
    csync;

    IMM32(I0,DCPLB_ADDR0);
    IMM32(P4,_b4_dcplb_address_pointer); /* save address pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I0;
    [P4] = R6;

    IMM32(I1,DCPLB_DATA0);
    IMM32(P4,_b4_dcplb_data_pointer); /* save data pointer for CPBL exception handler required for Round Robin Scheduling */
    R6 = I1;
    [P4] = R6;

    IMM32(I2,_b4_dcplbs_table);
    R5 = -1; /* end point comparison */
    R6 = 15; /* max counter */

    /* read entries from table */
    read_daddr:
        R7 = [I2++];
        CC = R7 == R5;
        if CC jump ddone;
        [I0++] = R7;
    read_ddata:
        R7 = [I2++];
        [I1++] = R7;
        R6 = R6 + R5;
        CC = R6 == R5;
        if !CC jump read_daddr;
    ddone:

    /* ENABLE DCPLBs */
    R7 = DMEM_CONTROL_VAL;
    [P5 + lo(DMEM_CONTROL)] = R7;
    ssync;


    dcplb_init.done:
    POP_DAG_REGx(2);
    POP_DAG_REGx(1);
    POP_DAG_REGx(0);
    (R7:5,P5:4) = [SP++];
    ASTAT = [SP++];
    unlink;


    rts;
__dcplb_init.END:


/****************************************************************************
 Function:          __dcplb_handler
 
 Description:       Identify correct DCPLB entry in DCPLB table
                    Replace one (Round Robin Scheduling) with the correct entry
 
 global variables:  _b4_dcplb_address_pointer
                    _b4_dcplb_data_pointer

 Registers used:    P5: DCPLB_FAULT_ADDR (fixed)
*****************************************************************************/


__dcplb_handler:


    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:1,P5:3);
    PUSH_DAG_REGx(0);
    PUSH_DAG_REGx(1);
    IMM32(P5,SYS_MMR_BASE);


    /**********************
    * Verify a DCPLB miss *
    **********************/

    R7 = SEQSTAT;
    R6 = 0x3F;
    R7 = R7 & R6;
    R6 = 0x26; /* Data access CPLB miss */
    CC = R7 == R6;
    if !CC jump dcplb_handler.done;


    /********************************************
    * Load starting point in _b4_dcplbs_table for: *
    * DCPLB_ADDR -> I0                          *
    * DCPLB_DATA -> I1                          *
    ********************************************/

    IMM32(I0,_b4_dcplbs_table); /* DCPLB_ADDRx */
    B0 = I0;
    M0 = 4;
    L0 = 8*dcplbs_table_length;

    
    /**************************************************
    * Search for the right entry in the _b4_dcplbs_table *
    * depending on DCPLB_FAULT_ADDR                   *
    **************************************************/

    R6 = [P5 + lo(DCPLB_FAULT_ADDR)];      /* DCPLB_FAULT_ADDR */
    IMM32(R1,0x00030000); /* for PAGE_SIZE bits */

    dcplb_handler.find_DCPLB_entry:
    R5 = [I0++];    /* DCPLB_ADDRx */
    R4 = [I0++];    /* DCPLB_DATAx */
    /* Extract PAGE_SIZE bits */
    R3 = R4 & R1;
    R3 = R4 >> 16;  /* R3 = CPLB Page Size bits */


    /*****************
    * find PAGE_SIZE *
    *****************/

    CC = R3 == 0x3;
    if CC jump MB4 (BP);    /* PAGE_SIZE = 4MB */
    CC = R3 == 0x2;
    if CC jump MB1;         /* PAGE_SIZE = 1MB */
    CC = R3 == 0x1;
    if CC jump kB4;         /* PAGE_SIZE = 4kB */
                            /* else: PAGE_SIZE = 1kB */

    kB1:                    /* 1kB: 0x0400 */
    R2 = 0x400;
    jump dcplb_handler.page_size_End;

    kB4:                    /* 4kB: 0x1000 */
    R2 = 0x1000;
    jump dcplb_handler.page_size_End;

    MB1:                    /* 1MB: 0x100000 */
    IMM32(R2,0x00100000);
    jump dcplb_handler.page_size_End;

    MB4:                    /* 4MB: 0x400000 */
    IMM32(R2,0x00400000);

    dcplb_handler.page_size_End:
    R7 = R5 + R2;           /* Store DCPLB_ADDRx end address */


    /***********************************************************
    * Check if actual entry in _b4_dcplbs_table is valid          *
    * DCPLB_ADDRx_START <= DCPLB_FAULT_ADDR < DCPLB_ADDRx_END  *
    * R5 <= R6 < R7                                            *
    ***********************************************************/

    CC = R5 <= R6;
    if !CC jump dcplb_handler.find_DCPLB_entry (BP);
    CC = R6 < R7;
    if !CC jump dcplb_handler.find_DCPLB_entry (BP);


    /**********************************************
    * Load last index point in _b4_dcplbs_table for: *
    * _b4_dcplb_address_pointer -> I2                 *
    * _b4_dcplb_data_pointer -> I3                    *
    * both need to be updated at the end          *
    * initialization in: __dcplb_init              *
    **********************************************/

    IMM32(B0,DCPLB_ADDR0); /* 0xFFE0 0200 */
    IMM32(P4,_b4_dcplb_address_pointer);
    R7 = [P4];
    I0 = R7;
    M0 = 4;                     /* 4 byte steps */
    L0 = 64;                    /* 16 DCPLB_ADDRx entrties with each 4 bytes */

    IMM32(B1,DCPLB_DATA0); /* 0xFFE0 0100 */
    IMM32(P3,_b4_dcplb_data_pointer);
    R7 = [P3];
    I1 = R7;
    M1 = M0;                    /* 4 byte steps */
    L1 = L0;                    /* 16 DCPLB_DATAx entrties with each 4 bytes */


    /****************************************
    * Check if actual DCPLB entry is locked *
    * If yes than try next one              *
    ****************************************/

    dcplb_handler.check_DCPLB_LOCK:
    R2 = [I0++];                /* DCPLB_ADDRx */
    NOP;
    R3 = [I1++];                /* DCPLB_DATAx */
    CC = bittst(R3,1);
    if CC jump dcplb_handler.check_DCPLB_LOCK;
    

    /* decrement index because the pointer is already at the next entry */
    R2 = [I0--];                /* DCPLB_ADDRx */
    R3 = [I1--];                /* DCPLB_DATAx */


    /**********************
    * Replace DCPLB entry *
    **********************/

    /* DISABLE DCPLBs to be able to update them */
    R7 = [P5 + lo(DMEM_CONTROL)];
    bitclr (R7,1);
    [P5 + lo(DMEM_CONTROL)] = R7;
    ssync;

    /* Update DCPLB entry and store address in _b4_dcplb_address_pointer and _b4_dcplb_data_pointer for Round Robin scheduling */
    [I0++] = R5;
    R7 = I0;
    [P4] = R7;
    [I1++] = R4;
    R7 = I1;
    [P3] = R7;

    /* ENABLE DCPLBs */
    R7 = DMEM_CONTROL_VAL;
    [P5 + lo(DMEM_CONTROL)] = R7;
    ssync;


    IMM32(P5,IMASK);
    R7 = [P5];
    bitclr (R7,bitpos(EVT_IVG14));
    [P5] = R7;


    dcplb_handler.done:
    POP_DAG_REGx(1);
    POP_DAG_REGx(0);
    (R7:1,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;


    rti;
__dcplb_handler.end:


#endif /* defined DATA_CACHE_EN */


/*****************************************************************************
 EOF
******************************************************************************/
