/*****************************************************************************
**                                                                          **
**   Name:  BF548 Host Boot BF54x                                           **
**                                                                          **
******************************************************************************

(C) Copyright 2008 - Analog Devices, Inc.  All rights reserved.

Project Name:   BF548_Host_Boot_BF54x

Date Modified:  05/07/08    v1.0    andreas pellkofer

History:        01/22/08    v0      andreas pellkofer

Tested
Software:       VisualDSP++ 5.0.3.0 (Update 3)

Supported
Software:       VisualDSP++ 5.0.3.0 (Update 3) and above

Tested
Hardware:       ADSP-BF548 EZ-KIT Lite Rev. 1.3 Silicon Rev. 0.1

Supported
Plattforms:     ADSP-BF548 EZ-KIT Lite

Connections:    ADDS-HPUSB-ICE

Settings:       Default

Purpose/
Description:    CHECK file 'configurations.h' first.
                All necessary system settings are done and described in there.
                This programm consists of three subparts for doing HOST boot
                operations for SPI, TWI & UART. These parts are embedded in a
                global routine expecting a PINT interrupt (PUSHBUTTON) to start
                one of the available boot modes.
                SPI Host Boot -> PB1, LED1
                TWI Host Boot -> PB2, LED2
                UART Host Boot -> PB3, LED3
                UART Host Boot with DMA -> PB4, LED4
                The core timer is showing via LED5 that there is still activity

                Alternatively one of the subparts can be called directly.
*****************************************************************************/


/****************************************************************************
 Include Section
*****************************************************************************/

#include "global.h"

#include "TWI.h"
#include "SPI.h"
#include "UART.h"

#include "PINT.h"
#include "CTIMER.h"

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

/*****************************************************************************
 Global Variables
******************************************************************************/

.GLOBAL loader_file;

/*****************************************************************************
 Variables
******************************************************************************/

.SECTION sdram;
.BYTE loader_file[] = INCLUDE_FILE;

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL _main;

/****************************************************************************
 Main Program
*****************************************************************************/

.SECTION program;

_main:

    call __clear_reg;
    call __led_setup;

    /* Either: */
    call __pint_run;
    call __ctimer;

    /* Or one of the following calls */
//    call __spi_hostboot_run;
//    call __twi_hostboot_run;
//    call __uart_hostboot_run;
//    call __uart_hostboot_dma_run;

_main.end: jump _main.end;

/*****************************************************************************
 EOF
******************************************************************************/
